#include <stdlib.h>
#include <string.h>
#include "matrix.h"

static int _mx_num(matrix *dst, double num, short mode);
static int _mx_add_matrix(matrix *dst, const matrix *a, const matrix *b, short mode);
static int _mx_set(const double *src, matrix *mx, unsigned short r, unsigned short c, short mode);

matrix
*mx_new()
{
	matrix *mx;

	mx = malloc(sizeof(*mx));
	if (mx != NULL) {
		mx->rows = 0;
		mx->cols = 0;
		mx->val = NULL;
	}

	return mx;
}

int
mx_free(matrix **mx)
{
	if (mx == NULL) {
		return MX_USERR;
	}

	if (*mx == NULL) {
		return MX_UNSET;
	}

	if ((*mx)->val != NULL) {
		free((*mx)->val);
	}

	free(*mx);
	(*mx) = NULL;

	return MX_OK;
}

int
mx_state(const matrix *mx)
{
	if (mx == NULL) {
		return MX_UNSET;
	}

	if (mx->val == NULL) {
		return MX_NOVAL;
	}

	return MX_OK;
}

int
mx_set(const double *src, matrix *mx, unsigned short rows, unsigned short cols)
{
	return _mx_set(src, mx, rows, cols, 0);
}

int mx_set_row(const double *src, matrix *mx, unsigned short r)
{
	return _mx_set(src, mx, r, 0, 1);
}

int
mx_set_col(const double *src, matrix *mx, unsigned short c)
{
	return _mx_set(src, mx, 0, c, 2);
}

static int
_mx_set(const double *src, matrix *mx, unsigned short r, unsigned short c, short mode)
{
	int i, res;

	res = mx_state(mx);
	if (res != MX_OK) {
		if (mode != 0) {
			return res;
		}
	}

	if (src == NULL) {
		return MX_USERR;
	}

	if (mode == 0) {
		res = mx_resize(mx, r, c);
		if (res != MX_OK) {
			return res;
		}

		memcpy(mx->val, src, r * c * sizeof(*src));
	}

	if (mode == 1) {
		if (r >= mx->rows) {
			return MX_USERR;
		}

		for (i = 0 ; i < mx->cols; i++) {
			mx->val[mx->cols*r + i] = src[i];
		}
	}
	
	if (mode == 2) {
		if (c >= mx->rows) {
			return MX_USERR;
		}

		for (i = 0 ; i < mx->rows; i++) {
			mx->val[mx->cols*i + c] = src[i];
		}
	}

	return MX_OK;
}

int
mx_set_cell(double val, matrix *mx, unsigned short row, unsigned short col)
{
	int res;

	res = mx_state(mx);

	if (res == MX_OK) {
		return res;
	}

	if (row >= mx->rows) {
		return MX_USERR;
	}

	if (col >= mx->cols) {
		return MX_USERR;
	}

	mx->val[row * mx->cols + col] = val;

	return MX_OK;
}

int mx_get_row(double *dst, matrix *mx, unsigned short r)
{
	int i, res;

	res = mx_state(mx);
	if (res != MX_OK) {
		return res;
	}

	if (dst == NULL) {
		return MX_USERR;
	}

	if (r >= mx->rows) {
		return MX_USERR;
	}

	for (i = 0 ; i < mx->cols; i++) {
		dst[i] = mx->val[mx->cols*r + i];
	}

	return MX_OK;
}

int
mx_get_col(double *dst, matrix *mx, unsigned short c)
{
	int i, res;

	res = mx_state(mx);
	if (res != MX_OK) {
		return res;
	}

	if (dst == NULL) {
		return MX_USERR;
	}

	if (c >= mx->rows) {
		return MX_USERR;
	}

	for (i = 0 ; i < mx->rows; i++) {
		dst[i] = mx->val[mx->cols*i + c];
	}

	return MX_OK;
}

int
mx_get_cell(double *dst, const matrix *mx, unsigned short row, unsigned short col)
{
	int res;

	if (dst == NULL) {
		return MX_USERR;
	}

	res = mx_state(mx);
	if (res != MX_OK) {
		return res;
	}

	if (row >= mx->rows) {
		return MX_USERR;
	}

	if (col >= mx->cols) {
		return MX_USERR;
	}

	*dst = mx->val[row * mx->cols + col];

	return MX_OK;
}

int
mx_swap_rows(matrix *mx, unsigned short r1, unsigned short r2)
{
	int res, i, tmp;

	res = mx_state(mx);
	if (res != MX_OK) {
		return res;
	}

	if (r1 >= mx->rows) {
		return MX_USERR;
	}

	if (r2 >= mx->rows) {
		return MX_USERR;
	}

	for (i = 0; i < mx->cols; i++) {
		tmp = mx->val[r1 * mx->cols + i];
		mx->val[r1 * mx->cols + i] = mx->val[r2 * mx->cols + i];
		mx->val[r2 * mx->cols + i] = tmp;
	}

	return MX_OK;
}

int
mx_swap_cols(matrix *mx, unsigned short c1, unsigned short c2)
{
	int res, i, tmp;

	res = mx_state(mx);
	if (res != MX_OK) {
		return res;
	}

	if (c1 >= mx->rows) {
		return MX_USERR;
	}

	if (c2 >= mx->rows) {
		return MX_USERR;
	}

	for (i = 0; i < mx->rows; i++) {
		tmp = mx->val[i * mx->cols + c1];
		mx->val[i * mx->cols + c1] = mx->val[i * mx->cols + c2];
		mx->val[i * mx->cols + c2] = tmp;
	}

	return MX_OK;
}

int
mx_resize(matrix *mx, unsigned short rows, unsigned short cols)
{
	double *tmp;
	size_t n;

	if (mx == NULL) {
		return MX_UNSET;
	}

	n = rows * cols;

	tmp = realloc(mx->val, n * sizeof(double));
	
	if (tmp == NULL) {
		return MX_NOMEM;
	}

	mx->val = tmp;
	mx->rows = rows;
	mx->cols = cols;

	mx_clear(mx);

	return MX_OK;
}

int
mx_transpose(matrix *mx)
{
	double *tmp;
	size_t n;
	int res, i, j;

	res = mx_state(mx);
	if (res != MX_OK) {
		return res;
	}

	n = mx->rows * mx->cols;

	tmp = malloc(n * sizeof(double));

	for (i = 0; i < mx->rows; i++) {
		for (j = 0; j < mx->cols; j++) {
			tmp[mx->rows * j + i] = mx->val[mx->cols * i + j];
		}
	}

	res = mx_set(tmp, mx, mx->cols, mx->rows);

	return res;
}

int
mx_copy(matrix *dst, const matrix *src)
{
	if (src == NULL) {
		return MX_UNSET;
	}

	if (dst == NULL) {
		return MX_USERR;
	}

	return mx_set(src->val, dst, src->rows, src->cols);
}

int
mx_rsign(matrix *mx)
{
	int i, res;
	size_t n;

	res = mx_state (mx);
	if (res != MX_OK) {
		return res;
	}

	n = mx->rows * mx->cols;

	for (i = 0; i < n; i++) {
		mx->val[i] = -mx->val[i]; 
	}

	return MX_OK;
}

int
mx_clear(matrix *mx)
{
	int res;

	res = mx_state(mx);

	if (res != MX_OK) {
		return res;
	}

	memset(mx->val, 0, mx->rows * mx->cols * sizeof(double));

	return MX_OK;
}

int
mx_add_num(matrix *dst, double num)
{
	return _mx_num(dst, num, 0);
}

int
mx_mult_num(matrix *dst, double num)
{
	return _mx_num(dst, num, 1);
}

static int
_mx_num(matrix *dst, double num, short mode)
{
	size_t n;
	int i, res;

	res = mx_state(dst);
	if (res != MX_OK) {
		return res;
	}

	n = dst->rows * dst->cols;

	if (mode == 0) {
		for (i = 0; i < n; i++) {
			dst->val[i] += num;
		}
	}

	if (mode == 1) {
		for (i = 0; i < n; i++) {
			dst->val[i] *= num;
		}
	}

	return MX_OK;
}

int mx_row_add_num(matrix *mx, unsigned short r, double num)
{
	int res, i;

	res = mx_state(mx);
	if (res != MX_OK) {
		return res;
	}

	if (r >= mx->rows) {
		return MX_USERR;
	}

	for (i = 0; i < mx->cols; i++) {
		res = r * mx->cols + i;
		mx->val[res] += num;
	}

	return MX_OK;
}

int mx_row_mult_num(matrix *mx, unsigned short r, double num)
{
	int res, i;

	res = mx_state(mx);
	if (res != MX_OK) {
		return res;
	}

	if (r >= mx->rows) {
		return MX_USERR;
	}

	for (i = 0; i < mx->cols; i++) {
		res = r * mx->cols + i;
		mx->val[res] *= num;
	}

	return MX_OK;
}

int
mx_col_add_num(matrix *mx, unsigned short c, double num)
{
	int res, i;

	res = mx_state(mx);
	if (res != MX_OK) {
		return res;
	}

	if (c >= mx->cols) {
		return MX_USERR;
	}

	for (i = 0; i < mx->cols; i++) {
		res = i * mx->cols + c;
		mx->val[res] += num;
	}

	return MX_OK;
}

int
mx_col_mult_num(matrix *mx, unsigned short c, double num)
{
	int res, i;

	res = mx_state(mx);
	if (res != MX_OK) {
		return res;
	}

	if (c >= mx->cols) {
		return MX_USERR;
	}

	for (i = 0; i < mx->cols; i++) {
		res = i * mx->cols + c;
		mx->val[res] *= num;
	}

	return MX_OK;
}

int
mx_add_matrix(matrix *dst, const matrix *a, const matrix *b)
{
	return _mx_add_matrix(dst, a, b, 0);
}

int
mx_sub_matrix(matrix *dst, const matrix *a, const matrix *b)
{
	return _mx_add_matrix(dst, a, b, 1);
}

static int
_mx_add_matrix(matrix *dst, const matrix *a, const matrix *b, short mode)
{
	size_t n;
	int i, res;

	res = mx_state(a);
	if (res != MX_OK) {
		return res;
	}

	res = mx_state(b);
	if (res != MX_OK) {
		return res;
	}

	if (a->rows != b->rows) {
		return MX_USERR;
	}

	if (a->cols != b->cols) {
		return MX_USERR;
	}

	res = mx_resize(dst, a->rows, b->cols);
	if (res != MX_OK) {
		return res;
	}

	n = dst->rows * dst->cols;

	if (mode == 0) {
		for (i = 0; i < n; i++) {
			dst->val[i] = a->val[i] + b->val[i];
		}
	}

	if (mode == 1) {
		for (i = 0; i < n; i++) {
			dst->val[i] = a->val[i] - b->val[i];
		}
	}

	return MX_OK;
}

int
mx_mult_matrix(matrix *dst, const matrix *a, const matrix *b)
{
	int i, j, k, target, a_pos, b_pos;
	int res;

	res = mx_state(a);
	if (res != MX_OK) {
		return res;
	}

	res = mx_state(b);
	if (res != MX_OK) {
		return res;
	}

	if (a->cols != b->rows) {
		return MX_USERR;
	}

	res = mx_resize(dst, a->rows, b->cols);
	if (res != MX_OK) {
		return res;
	}

	for (i = 0; i < b->rows; i++) {
		for (j = 0; j < b->cols; j++) {
			for (k = 0; k < b->rows; k++) {
				target = i * dst->cols + j;
				a_pos = i * a->cols + k;
				b_pos = b->rows * k + j;
				dst->val[target] += a->val[a_pos] * b->val[b_pos];
			}
		}
	}

	return MX_OK;
}
